# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                                                         #
#   Copyright 2014 by Francisco Pinto <francisco.pinto@epfl.ch>           #
#                                                                         #
#   Center for Digital Education (CEDE)                                   #
#   Ecole Polytecnique Federale de Lausanne (EPFL)                        #
#   1015 Lausanne, Switzerland                                            #
#                                                                         #
#   This program is free software: you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation, either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   Visit the GNU General Public License website for more details:        #
#   http://www.gnu.org/licenses/                                          #
#                                                                         #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


import xlrd, csv, pprint, re, gzip, gspread, json, hashlib, os, time, copy, sys, xmltodict
import simplejson as json
from termcolor import colored, cprint
import mysql.connector
from oauth2client.client import SignedJwtAssertionCredentials

from json import loads, dumps
from collections import OrderedDict

SafeTypeCodebook = {
	"'" : '<squote/>',
	'"' : '<dquote/>',
	'?' : '<intermark/>',
	'!' : '<exclmark/>',
	',' : '<comma/>',
	':' : '<colon/>',
	';' : '<semicolon/>',
	'(' : '<lpar/>',
	')' : '<rpar/>',
	'}' : '<lbrace/>',
	'{' : '<rbrace/>',
	'[' : '<lbracket/>',
	']' : '<rbracket/>',
	'\t' : '<tab/>',
	'\n' : '<linebreak/>',
	'\r' : '<linebreak/>',
}

# Pretty-print JSON
def printjson(s, Color='grey'):
	cprint(json.dumps(s, sort_keys=True, indent=4, separators=(',', ': ')), Color)

def issafetype(Input):
	for Key in SafeTypeCodebook.keys():
		if Key in Input:
			return False
	return True

def safetype(Input):
	if type(Input) is str or type(Input) is unicode:
		for Key in SafeTypeCodebook.keys():
			Input = Input.replace(Key, SafeTypeCodebook[Key])
	return Input

def unsafetype(Input, Exceptions=[]):
	if type(Input) is str or type(Input) is unicode:
		for Key in SafeTypeCodebook.keys():
			if not Key in Exceptions:
				Input = Input.replace(SafeTypeCodebook[Key], Key)
	return Input

def remove_parasite_chacters(String, Parasites, Replacements):
	ParDepth = 0
	SplitChars = list(String)
	for c in range(len(SplitChars)):
		if SplitChars[c]==Parasites[0]:
			ParDepth += 1
			if ParDepth>1:
				SplitChars[c] = Replacements[0]
		elif SplitChars[c]==Parasites[1]:
			if ParDepth>1:
				SplitChars[c] = Replacements[1]
			ParDepth -= 1
	return ''.join(SplitChars)
	

class LogFile:
	
	def __init__(self, FilePath=None, OAuthFile=None, Username=None, Password=None):

		if FilePath==None:
			self.DictList = None
			self.FilePath = None
		else:
			self.DictList = self.toDict(FilePath, OAuthFile, Username, Password)
			self.FilePath = FilePath

	def setDictList(self, NewDictList):
		self.DictList = NewDictList
		return

	def getDictList(self):
		return self.DictList

	def setFilePath(self, NewFilePath):
		self.FilePath = NewFilePath
		return

	def getFilePath(self):
		return self.FilePath

	def size(self):
		return len(self.DictList)

	#=========================================================================================================
	#===========================   Load Arbitrary Log File into Python DICT Type   ===========================
	#=========================================================================================================
	def toDict(self, FilePath, OAuthFile=None, Username=None, Password=None):

		#======================#
		# --- LOG Log Type --- #
		#======================#
		if FilePath.endswith('.log') or FilePath.endswith('.log.gz') or FilePath.endswith('.plog') or FilePath.endswith('.plog.gz') or FilePath.endswith('.mongo'):

			if FilePath.endswith('.log.gz') or FilePath.endswith('.plog.gz'):
				f = gzip.open(FilePath,'r')
			else:
				f = open(FilePath,'r')
			JSONBody = f.read()
			f.close()

			self.DictList = []
			for JSONItem in JSONBody.splitlines():
				self.DictList.append(json.loads(JSONItem))

		# END ELIF ----------

		#=======================#
		# --- JSON Log Type --- #
		#=======================#
		elif FilePath.endswith('.json') or FilePath.endswith('.json.gz'):

			if FilePath.endswith('.json.gz'):
				f = gzip.open(FilePath,'rb')
			else:
				f = open(FilePath,'rb')
			Body = f.read()
			f.close()

			self.DictList = [json.loads(Body)]
			
		# END ELIF ----------

		#=======================#
		# --- XLSX Log Type --- #
		#=======================#
		elif FilePath.endswith('.xlsx'):

			# Load workbook and worksheet
			workbook = xlrd.open_workbook(FilePath)
			worksheet = workbook.sheet_by_name(workbook.sheet_names()[0])

			# Get number of rows and cells
			NRows = worksheet.nrows - 1
			NCells = worksheet.ncols - 1

			# Get key names
			KeyNames = []
			for c in range(NCells):
				# Cell Types: 0=Empty, 1=Text, 2=Number, 3=Date, 4=Boolean, 5=Error, 6=Blank
				CellType = worksheet.cell_type(0, c)
				CellValue = worksheet.cell_value(0, c)
				KeyNames.append(safetype(CellValue))

			# Fill up dictionary
			self.DictList = []
			for r in range(1,NRows+1):
				row = worksheet.row(r)
				DictItem = dict()
				for c in range(NCells+1):
					CellType = worksheet.cell_type(r, c)
					CellValue = worksheet.cell_value(r, c)
					DictItem.update({KeyNames[c] : safetype(CellValue)})
				self.DictList.append(DictItem)

		# END ELIF ----------

		#======================#
		# --- CSV Log Type --- #
		#======================#
		elif FilePath.endswith('.csv') or FilePath.endswith('.csv.gz') or FilePath.endswith('.tcsv') or FilePath.endswith('.tcsv.gz'):

			csv.field_size_limit(sys.maxsize)

			if FilePath.endswith('.csv.gz') or FilePath.endswith('.tcsv.gz'):
				f = gzip.open(FilePath,'rb')
			else:
				f = open(FilePath,'rb')
			Body = f.read()
			f.close()

			# Determine CSV delimiter
			PotentialDelimitersCount = {
				';' : Body.count(';'),
				',' : Body.count(','),
				'|' : Body.count('|'),
				'\t' : Body.count('\t')
			}
			MaxCount = 0
			for PotentialDelimiter in PotentialDelimitersCount.keys():
				if PotentialDelimitersCount[PotentialDelimiter]>MaxCount:
					MaxCount = PotentialDelimitersCount[PotentialDelimiter]
					Delimiter = PotentialDelimiter
			
			CSVTable = csv.reader(Body.splitlines(), delimiter=Delimiter)
			CSVTable = list(CSVTable)

			# Transpose if needed
			if FilePath.endswith('.tcsv') or FilePath.endswith('.tcsv.gz'):
				CSVTable = map(list, zip(*CSVTable))

			# Convert table list to dict
			Keys = CSVTable[0]
			Rows = CSVTable[1:]
			self.DictList = []
			for row in Rows:
				self.DictList.append( dict(zip(Keys,row)) )

		# END ELIF ----------
		
		#=======================#
		# --- GDoc Log Type --- #
		#=======================#
		elif FilePath.startswith('docs.google.com'):

			if OAuthFile==None:
				print 'Please specify OAuth filepath using OAuthFile=\'...\' argument.'
				return

			Arguments = FilePath.split('/')

			DocumentName = Arguments[1]
			SheetName = Arguments[2]

			json_key = json.load(open(OAuthFile))
			scope = ['https://spreadsheets.google.com/feeds']

			credentials = SignedJwtAssertionCredentials(json_key['client_email'], json_key['private_key'], scope)
			GSpreadObj = gspread.authorize(credentials)

			SpreadDoc = GSpreadObj.open(DocumentName)
			Worksheet_Ready = SpreadDoc.worksheet(SheetName)

			Body =  Worksheet_Ready.export(format='csv').read()

			# Determine CSV delimiter
			PotentialDelimitersCount = {
				';' : Body.count(';'),
				',' : Body.count(','),
				'|' : Body.count('|'),
				'\t' : Body.count('\t')
			}
			MaxCount = 0
			for PotentialDelimiter in PotentialDelimitersCount.keys():
				if PotentialDelimitersCount[PotentialDelimiter]>MaxCount:
					MaxCount = PotentialDelimitersCount[PotentialDelimiter]
					Delimiter = PotentialDelimiter
			
			CSVTable = csv.reader(Body.splitlines(), delimiter=Delimiter)
			CSVTable = list(CSVTable)

			# Transpose if needed
			if FilePath.endswith('.tcsv') or FilePath.endswith('.tcsv.gz'):
				CSVTable = map(list, zip(*CSVTable))

			# Convert table list to dict
			Keys = CSVTable[0]
			Rows = CSVTable[1:]
			self.DictList = []
			for row in Rows:
				self.DictList.append( dict(zip(Keys,row)) )

		# END ELIF ----------

		#======================#
		# --- SQL Log Type --- #
		#======================#
		elif FilePath.endswith('.xml') or FilePath.endswith('.xml.gz'):

			if FilePath.endswith('.xml.gz'):
				f = gzip.open(FilePath,'rb')
			else:
				f = open(FilePath,'rb')
			Body = f.read()
			f.close()

			self.DictList = [loads(dumps(xmltodict.parse(Body)))]

		#======================#
		# --- SQL Log Type --- #
		#======================#
		elif FilePath.endswith('.sql') or FilePath.endswith('.sql.gz'):

			# Generate random schema name
			EventID = hashlib.md5()
			EventID.update(FilePath)
			SchemaName = EventID.hexdigest()

			TempFileCreated = False
			if FilePath.endswith('.sql.gz'):
				if not os.path.isfile(FilePath[:-3]):
					os.system('gzip -c -d "%s" > %s.sql' % (FilePath, SchemaName))
					TempFileCreated = True
				FilePath = SchemaName+'.sql'

			print 'Creating schema...'
			if Password==None:
				os.system('mysql -u%s -e "DROP DATABASE IF EXISTS temp_%s;"' % (Username, SchemaName))
				os.system('mysql -u%s -e "CREATE DATABASE temp_%s;"' % (Username, SchemaName))
				print 'Importing data into database...'
				os.system('mysql -u%s --max_allowed_packet=2G --default_character_set utf8 temp_%s < "%s"' % (Username, SchemaName, FilePath))
				print 'Done.'
			else:
				os.system('mysql -u%s -p%s -e "DROP DATABASE IF EXISTS temp_%s;"' % (Username, Password, SchemaName))
				os.system('mysql -u%s -p%s -e "CREATE DATABASE temp_%s;"' % (Username, Password, SchemaName))
				print 'Importing data into database...'
				os.system('mysql -u%s -p%s --max_allowed_packet=2G --default_character_set utf8 temp_%s < "%s"' % (Username, Password, SchemaName, FilePath))
				print 'Done.'

			self.DictList = []

			try:
				MySQLConnector = mysql.connector.connect(user=Username, password=Password, host='localhost', database='temp_'+SchemaName)
				print 'Connected to MySQL database %s.' % SchemaName
			except:
				print 'ERROR in function "toDict": To import a SQL file requires a running MySQL server.'
				return

			MySQLCursor = MySQLConnector.cursor()
			MySQLCursor.execute("SHOW TABLES IN `temp_%s`;" % SchemaName)
			AllTables = MySQLCursor.fetchall()

			for TableName in [str(Table[0]) for Table in AllTables]:	

				MySQLCursor.execute("SHOW COLUMNS FROM `temp_%s`.`%s`;" % (SchemaName, TableName))
		 		AllColumns = MySQLCursor.fetchall()

		 		Keys = [str(Column[0]) for Column in AllColumns]

		 		MySQLCursor.execute("SELECT %s FROM `temp_%s`.`%s`;" % ('`'+'`,`'.join(Keys)+'`', SchemaName, TableName))
		 		print 'Fetching table '+TableName
				
				while True:
					#print 'Fetching row...'
					Row = MySQLCursor.fetchone()
					if Row is None:
						break
					else:
						DictItem = dict(zip(Keys,Row))
						DictItem.update({'TableName' : TableName})
						self.DictList.append(DictItem)

		 	print 'Closing connection...'
			MySQLConnector.close()
			print 'Done.'
			
			if Password==None:
				os.system('mysql -u%s -e "DROP DATABASE temp_%s"' % (Username, SchemaName))
			else:
				os.system('mysql -u%s -p%s -e "DROP DATABASE temp_%s"' % (Username, Password, SchemaName))

			if TempFileCreated:
				os.remove(SchemaName+'.sql')

		else:
			print 'Unknown log file extension.'
			return None

		# END ELIF ----------

		# Return log file in Python DICT type
		return self.DictList

	# END DEF toDict ----------

	#=========================================================================================================
	#=====================================   General Purpose Functions   =====================================
	#=========================================================================================================

	# Row-Print LogFile
	def rprint(self):
		for DictItem in self.DictList:
			print DictItem

	# Pretty-Print LogFile
	def pprint(self):
		for DictItem in self.DictList:
			printjson(DictItem)

	# Remove empty spaces from dictionary keys (useful to convert directly into SQL colum names)
	def RemoveSpacesFromKeys(self):
		KeyConversionDict = {}
		for Key in self.DictList[0].keys():
			KeyConversionDict.update({Key : Key.replace(' ','')})
		self.RenameKeys(KeyConversionDict)

	# Rename dictionary key names to other custom names
	def RenameKeys(self, KeyConversionDict):
		for key in KeyConversionDict:
			for k in range(len(self.DictList)):
				self.DictList[k][KeyConversionDict[key]] = self.DictList[k].pop(key)
		return self.DictList

	#=========================================================================================================
	#===================================   Save LogFile as JSON LOG File   ===================================
	#=========================================================================================================
	def toLOGFile(self, FilePath):

		LOGString = ''
		for DictItem in self.DictList:
			try:
				LOGString += json.dumps(DictItem)+'\n'
			except Exception, e:
				if len(re.findall(r'bytearray\(.*\) is not JSON serializable', str(e))):
					for k in DictItem:
						if type(DictItem[k]) is bytearray:
							DictItem[k] = str(DictItem[k])
					LOGString += json.dumps(DictItem)+'\n'

		# END FOR ----------

		if FilePath.endswith('.log.gz') or FilePath.endswith('.plog.gz'):
			f = gzip.open(FilePath, 'w')
		elif FilePath.endswith('.log') or FilePath.endswith('.plog'):
			f = open(FilePath, 'w')
		else:
			print 'Wrong file type'
			return
		f.write(LOGString)
		f.close()

	# END DEF toLOGFile ----------

	#=========================================================================================================
	#==================================   Save LogFile as SQL Query File   ===================================
	#=========================================================================================================
	def toSQLFile(self, FilePath, SchemaName, UniqueTableName=None, DataTypes=None, IncludeDropTableSQL=False, IncludeCreateTableSQL=True, OverwriteDuplicates=False):

		# Convert Python DICT Type to SQL "CREATE TABLE" and "INSERT INTO" Commands
		SQLInserts = self.toSQL(SchemaName, UniqueTableName, DataTypes, IncludeDropTableSQL, IncludeCreateTableSQL, OverwriteDuplicates)

		# Open new .sql or .sql.gz file
		if FilePath.endswith('.sql.gz'):
			f = gzip.open(FilePath, 'w')
		elif FilePath.endswith('.sql'):
			f = open(FilePath, 'w')
		else:
			print 'Wrong file type in function "toSQLFile".'
			return

		# Save SQL Query to file and close
		f.write(SQLInserts.encode('utf8'))
		f.close()

		# Return success
		return True

	# END DEF toSQLFile ----------

	#=========================================================================================================
	#=============   Convert Python DICT Type to SQL "CREATE TABLE" and "INSERT INTO" Commands   =============
	#=========================================================================================================
	def toSQL(self, SchemaName, UniqueTableName=None, DataTypes=None, IncludeDropTableSQL=False, IncludeCreateTableSQL=True, OverwriteDuplicates=False):

		# Get all table names in Log File
		TableNamesAndColumns = set()
		for DictItem in self.DictList:
			if UniqueTableName==None:
				TableName = DictItem['TableName']
			else:
				TableName = UniqueTableName
			TableNamesAndColumns.add(tuple([TableName])+tuple(DictItem.keys()))
		
		# Initialize global SQL Query
		SQLQuery = ''


		#--------------------------------------
		# Generate all "CREATE TABLE" commands
		#--------------------------------------

		if IncludeCreateTableSQL:

			# Loop though all items of the type {TableName, Column1, Column2, ...}
			for Item in TableNamesAndColumns:
				
				# Get table name
				TableName = Item[0]

				# If flag is active, generate "DROP TABLE" commands
				if IncludeDropTableSQL:
					SQLQuery_DropTable = "DROP TABLE IF EXISTS `%s`.`%s`;" % (SchemaName, TableName)
					SQLQuery += SQLQuery_DropTable+'\n'

				# Get all column names
				ColumnNames = Item[1:]

				# SQL Query Construction: CREATE TABLE IF NOT EXISTS SchemaName.TableName (
				SQLQuery_CreateTable = "CREATE TABLE IF NOT EXISTS `%s`.`%s` (" % (SchemaName, TableName)

				# SQL Query Construction: CREATE TABLE IF NOT EXISTS SchemaName.TableName (Column1 DataType1, Column2 DataType2, ...
				if DataTypes==None:
					for ColumnName in ColumnNames:
						# If data types not are given as argument, use LONGTEXT data type (highly inefficient)
			  			SQLQuery_CreateTable += "\n\t`%s` LONGTEXT DEFAULT NULL," % ColumnName
		  		else:	
					for ColumnName in DataTypes[TableName].keys():
						if ColumnName=='PrimaryKey':
							continue
						# Use custom data types if given as argument
						SQLQuery_CreateTable += "\n\t`%s` %s," % (ColumnName, DataTypes[TableName][ColumnName])
						
		  			# END FOR ----------

		  			# SQL Query Construction: CREATE TABLE IF NOT EXISTS SchemaName.TableName (Column1 DataType1, Column2 DataType2, ..., PRIMARY KEY (PrimaryKey)
		  			SQLQuery_CreateTable += "\nPRIMARY KEY (`%s`) " % DataTypes[TableName]['PrimaryKey']

		  		# SQL Query Construction: CREATE TABLE IF NOT EXISTS SchemaName.TableName (Column1 DataType1, Column2 DataType2, ..., PRIMARY KEY (PrimaryKey)) ENGINE=InnoDB DEFAULT CHARSET=utf8;
		  		SQLQuery_CreateTable = SQLQuery_CreateTable[:-1] + "\n) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
				
				# Add CREATE TABLE command to global SQL Query
				SQLQuery += SQLQuery_CreateTable+'\n'
			
			# END FOR ----------

		#-------------------------------------
		# Generate all "INSERT INTO" commands
		#-------------------------------------

		# Initialize List of INSERT commands
		SQLQuery_InsertValues_List = list()

		# Loop through all Dictionary rows
		for DictItem in self.DictList:

			# Get table name
			if UniqueTableName==None and ('TableName' in DictItem.keys()):
				TableName = DictItem['TableName']
			elif UniqueTableName!=None:
				TableName = UniqueTableName
			else:
				print 'Table name not found.'
				return

			# Get all columns names
			ColumnNames = DictItem.keys()

			# SQL Query Construction: INSERT INTO SchemaName.TableName (Column1, Column2, ...) VALUES (
			if OverwriteDuplicates:
				SQLQuery_InsertValues = "REPLACE"
			else:
				SQLQuery_InsertValues = "INSERT"
			SQLQuery_InsertValues += " INTO `%s`.`%s` (%s) VALUES (" % (SchemaName, TableName, '`'+'`, `'.join(ColumnNames)+'`')

			# SQL Query Construction: INSERT INTO SchemaName.TableName (Column1, Column2, ...) VALUES ( <-- next loop puts values here
			for ColumnName in ColumnNames:
				Value = DictItem[ColumnName]
				
				# SQL Query Construction: INSERT INTO SchemaName.TableName (Column1, Column2, ...) VALUES (Val1, Val2, ...
				# Handle STRINGs and NULLs
				if type(Value) is str or type(Value) is unicode:
					if Value=='' or Value=='NULL' or Value=='NaN':
						# Concatenate NULL type value
						SQLQuery_InsertValues += "NULL, "
					else:
						# Remove (problematic!) escape characters and quotes from strings
						Value = Value.replace("'", '<squote/>').replace("\\", '<backslash/>')
						# Concatenate STRING type value
 	  					SQLQuery_InsertValues += "'%s', " % Value
 	  			# Concatenate INT type values
 	  			elif type(Value) is int or (type(Value) is float and Value.is_integer()):
 	  				SQLQuery_InsertValues += "%d, " % int(Value)
 	  			# Concatenate FLOAT type value
 	  			elif type(Value) is float and not Value.is_integer():
 	  				SQLQuery_InsertValues += "%f, " % Value
 	  			# Concatenate BOOL type value
 	  			elif type(Value) is bool:
 	  				SQLQuery_InsertValues += "%s, " % Value
 	  			# Concatenate NULL if anything else fails
 	  			else:
 	  				SQLQuery_InsertValues += "NULL, "

 	  		# END FOR ----------

 	  		# SQL Query Construction: INSERT INTO SchemaName.TableName (Column1, Column2, ...) VALUES (Val1, Val2, ...) ON DUPLICATE KEY UPDATE PrimaryKey=PrimaryKey;
 	  		if OverwriteDuplicates:
 	  			SQLQuery_InsertValues = SQLQuery_InsertValues[:-2] + ");"
 	  		else:
 	  			if DataTypes!=None:
	  				SQLQuery_InsertValues = SQLQuery_InsertValues[:-2] + ") ON DUPLICATE KEY UPDATE `%s`=`%s`;" % (DataTypes[TableName]['PrimaryKey'], DataTypes[TableName]['PrimaryKey'])
				else:
					SQLQuery_InsertValues = SQLQuery_InsertValues[:-2] + ");"
			
			# Add SQL Query to List of INSERT commands
			SQLQuery_InsertValues_List.append( SQLQuery_InsertValues )

		# END FOR ----------

		# Break INSERT list into string and add to global SQL String
		SQLQuery += '\n'.join(SQLQuery_InsertValues_List)

		# Return global SQL Query
		return SQLQuery

	# END DEF toSQL ----------

#=========================
#== Main shell callback ==
#=========================
if __name__ == "__main__":

	print 'Nothing to do.'


